/* Project 4 by David Taralla | 23-02-2010
*  ========================================
*  This program executes a quaternionic rotation
*  of a vector around a given axis and display as
*  result the 3 components of the computed vector.
*/



/* New types used in Project 4
*  ============================
*  - Vector : floats x, y, z
*  - Quaternion : float angle, Vector vector
*
*  -> See "Functions on vectors" category below for
*     available operations on the Quaternion type.
*  -> See "Functions on quaternions" category below for
*     available operations on the Quaternion type.
*/
typedef struct {
    float x;
    float y;
    float z;
} Vector;

typedef struct {
    float angle;
    Vector vector;
} Quaternion;

#define _GNU_SOURCE // Needed to allow M_PI macro in math.h
#include <stdio.h>
#include <math.h>

//#include "prototypes.h"


/*************************************
*------------ Functions -------------*
*************************************/
// --- Functions on vectors ---
static float scalarProduct(Vector v1, Vector v2) {
    return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

static Vector vectorProduct(Vector v1, Vector v2) {
    Vector producedVector;
    
    producedVector.x = v1.y * v2.z - v1.z * v2.y;
    producedVector.y = v1.z * v2.x - v1.x * v2.z;
    producedVector.z = v1.x * v2.y - v1.y * v2.x;
    
    return producedVector;
}


// --- Functions on quaternions ---
/* --------------------------------------------------------
*  getQuatNorm description
*  ========================
*  Quaternion property : the square root of the sum of
*  of all components squares gives the norm of the
*  quaternion.
*  --------------------------------------------------------
*  getQuatNorm return value
*  =========================
*  This function returns the norm of [q].
*  --------------------------------------------------------
*/
static float getQuatNorm(Quaternion q) {
    return sqrt(q.angle * q.angle
              + q.vector.x * q.vector.x
              + q.vector.y * q.vector.y
              + q.vector.z * q.vector.z);
}


/* --------------------------------------------------------
*  getNormalizedQuat description
*  ==============================
*  In order to normalize a quaternion, we have to divide
*  it by its norm.
*  --------------------------------------------------------
*  getNormalizedQuat return value
*  ===============================
*  This function returns the normalized form of [q].
*  --------------------------------------------------------
*/
static Quaternion getNormalizedQuat(Quaternion q) {
    float quatNorm = getQuatNorm(q);
    
    q.angle /= quatNorm;
    q.vector.x /= quatNorm;
    q.vector.y /= quatNorm;
    q.vector.z /= quatNorm;
    
    return q;
}


/* --------------------------------------------------------
*  conjugateQuat description
*  ==========================
*  In order to conjugate a quaternion, we have to
*  keep its real part unchanged and take the
*  opposites of its imaginary parts.
*  --------------------------------------------------------
*  conjugateQuat return value
*  ===========================
*  This function returns the conjugated form of [q].
*  --------------------------------------------------------
*/
static Quaternion conjugateQuat(Quaternion q) {
    q.vector.x = -q.vector.x;
    q.vector.y = -q.vector.y;
    q.vector.z = -q.vector.z;
    
    return q;
}


/* --------------------------------------------------------
*  invertQuat description
*  =======================
*  In order to invert a quaternion, we have to divide
*  its conjugate by the square of its norm.
*  --------------------------------------------------------
*  invert Quatreturn value
*  ========================
*  This function returns the inverted form of [q].
*  --------------------------------------------------------
*/
static Quaternion invertQuat(Quaternion q) {
    q = conjugateQuat(q);
    float quatNorm = getQuatNorm(q);
    
    q.angle /= quatNorm * quatNorm;
    q.vector.x /= quatNorm * quatNorm;
    q.vector.y /= quatNorm * quatNorm;
    q.vector.z /= quatNorm * quatNorm;
    
    return q;
}


/* --------------------------------------------------------
*  multiplyQuat description
*  =========================
*  In order to multiply a quaternion by another,
*  we use theses formulas : if
*  Q1 = a + v1(b, c, d) and Q2 = w + v2(x, y, z)
*  then Q3 = Q1 x Q2 has these components :
*  Q3 = aw - scalarProduct(v1, v2)
*     + (av2 + wv1 + vectorProduct(v1, v2))
*  --------------------------------------------------------
*  multiplyQuat return value
*  ==========================
*  This function returns the Q3 quaternion.
*  --------------------------------------------------------
*/
static Quaternion multiplyQuat(Quaternion q1, Quaternion q2) {
    Quaternion multipliedQuat;
    
    multipliedQuat.angle = q1.angle * q2.angle
                         - scalarProduct(q1.vector, q2.vector);
    
    Vector q1q2Product = vectorProduct(q1.vector, q2.vector);
	
    multipliedQuat.vector.x = q1.angle * q2.vector.x
                            + q2.angle * q1.vector.x
                            + q1q2Product.x;
                            
    multipliedQuat.vector.y = q1.angle * q2.vector.y
                            + q2.angle * q1.vector.y
                            + q1q2Product.y;
							
    multipliedQuat.vector.z = q1.angle * q2.vector.z
                            + q2.angle * q1.vector.z
                            + q1q2Product.z;
    
    //multipliedQuat = getNormalizedQuat(multipliedQuat);
    return multipliedQuat;
}


// --- ! Kernel function ! ---
/* --------------------------------------------------------
*  rotateQuat description
*  =======================
*  Rotate of X degrees a given vector around a given axis.
*  For further details on the mathematical method used to
*  compute the vector, see the wikipedia article joint with
*  project 4 statement.
*  --------------------------------------------------------
*  rotateQuat return value
*  ========================
*  This function returns the solution vector asked for
*  Project 4.
*  --------------------------------------------------------
*/
static Vector rotateQuat(Quaternion rotationData, Vector axis) {
    Quaternion rotatedQuat;
    rotatedQuat.angle = cos(rotationData.angle / 2);
    
    float axisCoefficient = sin(rotationData.angle / 2);
    
    rotatedQuat.vector.x = axisCoefficient * axis.x;
    rotatedQuat.vector.y = axisCoefficient * axis.y;
    rotatedQuat.vector.z = axisCoefficient * axis.z;
    
    /* Now rotatedQuat is ready to be used in the formula
     * u X (vector to rotate) X (1/u) where u is rotatedQuat
     * and "X" means quaternion product (not commutative !).*/
    Quaternion vectToRotate = rotationData;
    vectToRotate.angle = 0; // This is only a vector, but quaternion
                            // operations are applied on it (see wiki).
    
    vectToRotate = multiplyQuat(rotatedQuat, vectToRotate);
    vectToRotate = multiplyQuat(vectToRotate, invertQuat(rotatedQuat));
    
    return vectToRotate.vector;
}


// --- Other functions ---
static float degreesToRad(float degrees) {
    return degrees * (M_PI / 180);
}


/*************************************
*----------- Main program -----------*
*************************************/
int main() {
    Quaternion rotationAxis;
    Quaternion rotationData;
    
    printf("Enter a vector (3 coefficients): ");
    if (scanf("%f %f %f", &rotationData.vector.x,    // Poor style but accepted
                          &rotationData.vector.y,    // by J. Piater while in C
                          &rotationData.vector.z) < 3) {
        
        printf("Invalid input\n");
        return 0;
    }
    
    printf("Enter a rotation axis (3 coefficients): ");
    if (scanf("%f %f %f", &rotationAxis.vector.x,    // Idem
                          &rotationAxis.vector.y,
                          &rotationAxis.vector.z) < 3) {
        
        printf("Invalid input\n");
        return 0;
    }
    rotationAxis = getNormalizedQuat(rotationAxis);
    
    printf("Enter a rotation angle (degrees): ");
    if (scanf("%f", &rotationData.angle) < 1){
        printf("Invalid input\n");
        return 0;
    }
    rotationData.angle = degreesToRad(rotationData.angle);
    
    // No invalid inputs, now computing and concluding
    Vector rotatedVect = rotateQuat(rotationData, rotationAxis.vector);
    printf("Rotated vector: %f %f %f\n", rotatedVect.x, rotatedVect.y,
                                         rotatedVect.z);
    
    return 0;
}
